#define _GNU_SOURCE
#include <inttypes.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <sys/mman.h>
#include "bbb_dmtimer.h"

#define AM335x_CM_PER       (uint32_t *)0x44E00000  // to 0x44E003FF
#define AM335x_CM_WKUP      (uint32_t *)0x44E00400  // to 0x44E004FF
#define AM335x_CM_DPLL      (uint32_t *)0x44E00500  // to 0x44E005FF
#define AM335x_CM_MPU       (uint32_t *)0x44E00600  // to 0x44E006FF
#define AM335x_CM_DEVICE    (uint32_t *)0x44E00700  // to 0x44E007FF
#define AM335x_CM_RTC       (uint32_t *)0x44E00800  // to 0x44E008FF
#define AM335x_CM_GFX       (uint32_t *)0x44E00900  // to 0x44E009FF
#define AM335x_CM_CEFUSE    (uint32_t *)0x44E00A00  // to 0x44E00AFF
#define AM335x_CM_PER_LEN   (size_t)0x600
#define AM335x_CM_DPLL_LEN  (size_t)0x100

#define AM335x_DMTIMER0     (uint32_t *)0x44E05000  // to 0x44E05FFF
#define AM335x_DMTIMER1MS   (uint32_t *)0x44E31000  // to 0x44E31FFF
#define AM335x_DMTIMER2     (uint32_t *)0x48040000  // to 0x48040FFF
#define AM335x_DMTIMER3     (uint32_t *)0x48042000  // to 0x48042FFF
#define AM335x_DMTIMER4     (uint32_t *)0x48044000  // to 0x48044FFF
#define AM335x_DMTIMER5     (uint32_t *)0x48046000  // to 0x48046FFF
#define AM335x_DMTIMER6     (uint32_t *)0x48048000  // to 0x48048FFF
#define AM335x_DMTIMER7     (uint32_t *)0x4804A000  // to 0x4804AFFF
#define AM335x_DMTIMER_LEN  (size_t)0x1000

_Bool bbb_dmtimer_init(void){
    int dev_map;
    if((dev_map = open("/dev/mem", O_RDWR|O_SYNC|O_LARGEFILE)) < 0){
        perror("open");
        return(1);
    }

    if((CM_PER = mmap(0, AM335x_CM_PER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_CM_PER)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map CM_PER registers. errno=%d\n",__func__, errno);
        return(1);
    }

    CM_DPLL=CM_PER+0x500/sizeof(uint32_t);

    if((DMTIMER[0] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER0)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER0 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[1] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER1MS)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER1MS registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[2] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER2)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER2 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[3] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER3)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER3 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[4] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER4)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER4 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[5] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER5)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER5 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[6] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER6)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER6 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    if((DMTIMER[7] = mmap(0, AM335x_DMTIMER_LEN, PROT_READ | PROT_WRITE, MAP_SHARED, dev_map, (off_t)AM335x_DMTIMER7)) == MAP_FAILED){
        fprintf(stderr, "Err: %s: Can not map DMTIMER7 registers. errno=%d\n",__func__, errno);
        return(1);
    }
    close(dev_map);
    return(0);
}

void bbb_dmtimer_free(void){
    if(CM_PER){
        munmap(CM_PER, AM335x_CM_PER_LEN);
        CM_PER = NULL;
        munmap(CM_DPLL, AM335x_CM_DPLL_LEN);
        CM_DPLL = NULL;
        for(uint8_t i=0; i<8; i++){
            munmap(DMTIMER[i], AM335x_DMTIMER_LEN);
            DMTIMER[i] = NULL;
        }
    }
}
