// CM registers
uint32_t *CM_PER;
enum cm_per_reg_index {
    CM_PER_L4LS_CLKSTCTRL,		// 0h
    CM_PER_L3S_CLKSTCTRL,		// 4h
    CM_PER_DUMMY0,		        // 8h Gap
    CM_PER_L3_CLKSTCTRL,		// Ch
    CM_PER_DUMMY1,		        // 10h Gap
    CM_PER_CPGMAC0_CLKCTRL,		// 14h
    CM_PER_LCDC_CLKCTRL,		// 18h
    CM_PER_USB0_CLKCTRL,		// 1Ch
    CM_PER_DUMMY2,		        // 20h Gap
    CM_PER_TPTC0_CLKCTRL,		// 24h
    CM_PER_EMIF_CLKCTRL,		// 28h
    CM_PER_OCMCRAM_CLKCTRL,		// 2Ch
    CM_PER_GPMC_CLKCTRL,		// 30h
    CM_PER_MCASP0_CLKCTRL,		// 34h
    CM_PER_UART5_CLKCTRL,		// 38h
    CM_PER_MMC0_CLKCTRL,		// 3Ch
    CM_PER_ELM_CLKCTRL,		    // 40h
    CM_PER_I2C2_CLKCTRL,		// 44h
    CM_PER_I2C1_CLKCTRL,		// 48h
    CM_PER_SPI0_CLKCTRL,		// 4Ch
    CM_PER_SPI1_CLKCTRL,		// 50h
    CM_PER_DUMMY3,		        // 54h Gap
    CM_PER_DUMMY4,		        // 58h Gap
    CM_PER_DUMMY5,		        // 5Ch Gap
    CM_PER_L4LS_CLKCTRL,		// 60h
    CM_PER_DUMMY6,		        // 64h Gap
    CM_PER_MCASP1_CLKCTRL,		// 68h
    CM_PER_UART1_CLKCTRL,		// 6Ch
    CM_PER_UART2_CLKCTRL,		// 70h
    CM_PER_UART3_CLKCTRL,		// 74h
    CM_PER_UART4_CLKCTRL,		// 78h
    CM_PER_TIMER7_CLKCTRL,		// 7Ch
    CM_PER_TIMER2_CLKCTRL,		// 80h
    CM_PER_TIMER3_CLKCTRL,		// 84h
    CM_PER_TIMER4_CLKCTRL,		// 88h
    CM_PER_DUMMY7,		        // 8Ch Gap
    CM_PER_DUMMY8,		        // 90h Gap
    CM_PER_DUMMY9,		        // 94h Gap
    CM_PER_DUMMY10,		        // 98h Gap
    CM_PER_DUMMY11,		        // 9Ch Gap
    CM_PER_DUMMY12,		        // A0h Gap
    CM_PER_DUMMY13,		        // A4h Gap
    CM_PER_DUMMY14,		        // A8h Gap
    CM_PER_GPIO1_CLKCTRL,		// ACh
    CM_PER_GPIO2_CLKCTRL,		// B0h
    CM_PER_GPIO3_CLKCTRL,		// B4h
    CM_PER_DUMMY15,		        // B8h Gap
    CM_PER_TPCC_CLKCTRL,		// BCh
    CM_PER_DCAN0_CLKCTRL,		// C0h
    CM_PER_DCAN1_CLKCTRL,		// C4h
    CM_PER_DUMMY16,		        // C8h Gap
    CM_PER_EPWMSS1_CLKCTRL,		// CCh
    CM_PER_DUMMY17,		        // D0h Gap
    CM_PER_EPWMSS0_CLKCTRL,		// D4h
    CM_PER_EPWMSS2_CLKCTRL,		// D8h
    CM_PER_L3_INSTR_CLKCTRL,	// DCh
    CM_PER_L3_CLKCTRL,		    // E0h
    CM_PER_IEEE5000_CLKCTRL,	// E4h
    CM_PER_PRU_ICSS_CLKCTRL,	// E8h
    CM_PER_TIMER5_CLKCTRL,		// ECh
    CM_PER_TIMER6_CLKCTRL,		// F0h
    CM_PER_MMC1_CLKCTRL,		// F4h
    CM_PER_MMC2_CLKCTRL,		// F8h
    CM_PER_TPTC1_CLKCTRL,		// FCh
    CM_PER_TPTC2_CLKCTRL,		// 100h
    CM_PER_DUMMY18,		        // 104h Gap
    CM_PER_DUMMY19,		        // 108h Gap
    CM_PER_SPINLOCK_CLKCTRL,	// 10Ch
    CM_PER_MAILBOX0_CLKCTRL,	// 110h
    CM_PER_DUMMY20,		        // 114h Gap
    CM_PER_DUMMY21,		        // 118h Gap
    CM_PER_L4HS_CLKSTCTRL,		// 11Ch
    CM_PER_L4HS_CLKCTRL,		// 120h
    CM_PER_DUMMY22,		        // 124h Gap
    CM_PER_DUMMY23,		        // 128h Gap
    CM_PER_OCPWP_L3_CLKSTCTRL,	// 12Ch
    CM_PER_OCPWP_CLKCTRL,		// 130h
    CM_PER_DUMMY24,		        // 134h Gap
    CM_PER_DUMMY25,		        // 138h Gap
    CM_PER_DUMMY26,		        // 13Ch Gap
    CM_PER_PRU_ICSS_CLKSTCTRL,	// 140h
    CM_PER_DUMMY27,		        // 144h Gap
    CM_PER_LCDC_CLKSTCTRL,		// 148h
    CM_PER_CLKDIV32K_CLKCTRL,	// 14Ch
    CM_PER_CLK_24MHZ_CLKSTCTRL 	// 150h
};
// CM_PER_TIMERx_CLKCTRL value
// 0x0 by default
#define IDLE_STATUS_MASK        0x00030000
#define MODULE_MODE_MASK        0x00000002
#define MODULE_DISABLED         0x0
#define MODULE_ENABLED          0x2


uint32_t *CM_DPLL;
enum cm_dpll_reg_index {
    CM_DPLL_DUMMY0,             // 0h Gap
	CLKSEL_TIMER7_CLK,		    // 4h
	CLKSEL_TIMER2_CLK,		    // 8h
	CLKSEL_TIMER3_CLK,		    // Ch
	CLKSEL_TIMER4_CLK,		    // 10h
	CM_MAC_CLKSEL,		        // 14h
	CLKSEL_TIMER5_CLK,		    // 18h
	CLKSEL_TIMER6_CLK,		    // 1Ch
	CM_CPTS_RFT_CLKSEL,		    // 20h
    CM_DPLL_DUMMY1,             // 24h Gap
	CLKSEL_TIMER1MS_CLK,		// 28h
	CLKSEL_GFX_FCLK,		    // 2Ch
	CLKSEL_PRU_ICSS_OCP_CLK,	// 30h
	CLKSEL_LCDC_PIXEL_CLK,		// 34h
	CLKSEL_WDT1_CLK,		    // 38h
	CLKSEL_GPIO0_DBCLK		    // 3Ch
};
// CLKSEL_TIMERx_CLK value
// 0x1 by default
#define TCLKIN                  0x0
#define CLK_M_OSC               0x1     // 24MHz
#define CLK_32KHz               0x2

// CLKSEL_TIMER1MS_CLK value
// 0x0 by default
#define CLK_M_OSC_1MS           0x0
#define CLK_32KHz_1MS           0x1
#define TCLKIN_1MS              0x2
#define CLK_RC32K_1MS           0x3
#define CLK32768_1MS            0x4

// TIMER registers
uint32_t *DMTIMER[8];
enum timer_reg_index {
    TIDR,		    // 0x0
    TM_DUMMY0, TM_DUMMY1, TM_DUMMY2,		// Gap
    TIOCP_CFG,		// 0x10
    TM_DUMMY3, TM_DUMMY4, TM_DUMMY5,		// Gap
    IRQ_EOI,		// 0x20
    IRQSTATUS_RAW,	// 0x24
    IRQSTATUS,		// 0x28
    IRQENABLE_SET,	// 0x2c
    IRQENABLE_CLR,	// 0x30
    IRQWAKEEN,		// 0x34
    TCLR,		    // 0x38
    TCRR,		    // 0x3c
    TLDR,		    // 0x40
    TTGR,		    // 0x44
    TWPS,		    // 0x48
    TMAR,		    // 0x4c
    TCAR1,		    // 0x50
    TSICR,		    // 0x54
    TCAR2		    // 0x58
};

// IRQSTATUS flags
#define TCAR_IT_FLAG        (1<<2)
#define OVF_IT_FLAG         (1<<1)
#define MAT_IT_FLAG         1

// IRQSTATUS flags
#define TCAR_EN_FLAG        (1<<2)
#define OVF_EN_FLAG         (1<<1)
#define MAT_EN_FLAG         1

_Bool bbb_dmtimer_init(void);
void bbb_dmtimer_free(void);
